
NAME = 'tickerdax'
URL = f'https://{NAME}.com'


class KeyTypes:
    REST = 'REST'
    WEBSOCKET = 'WEBSOCKET'


class Envs:
    DEV = f'{NAME.upper()}_DEV'
    OFFICIAL_DOCKER_IMAGE = f'{NAME.upper()}_OFFICIAL_DOCKER_IMAGE'
    EMAIL = f'{NAME.upper()}_EMAIL'
    REST_API_KEY = f'{NAME.upper()}_REST_API_KEY'
    WEBSOCKET_API_KEY = f'{NAME.upper()}_WEBSOCKET_API_KEY'
    CACHE_ROOT = f'{NAME.upper()}_CACHE_ROOT'
    CONFIG = f'{NAME.upper()}_CONFIG'
    REDIS_SERVER_ADDRESS = f'{NAME.upper()}_REDIS_SERVER_ADDRESS'
    REDIS_SERVER_PORT = f'{NAME.upper()}_REDIS_SERVER_PORT'



import os
import json
import subprocess
import sys
import asyncio
import time

import requests
import tempfile
import logging
import redis
import docker
import socket
import websockets
from websockets.exceptions import InvalidStatusCode, ConnectionClosed, ConnectionClosedOK
from logging.handlers import TimedRotatingFileHandler
from tickerdax.constants import Envs, KeyTypes, NAME, URL
from docker.errors import DockerException
from datetime import datetime, timedelta, timezone
from tickerdax import formatting
from pprint import pformat, pprint


class TickerDax:
    def __init__(
            self,
            email=None,
            rest_api_key=None,
            websocket_api_key=None,
            fast_start=True,
            connect=True,
            log_connection=True,
            force=False,
            disable_logging=False,
            debug=False
    ):
        websocket_api_prefix = 'ws'
        rest_api_prefix = 'api'
        if os.environ.get(Envs.DEV):
            rest_api_prefix = 'dev-api'
            websocket_api_prefix = 'dev-ws'

        # general configuration
        self.supported_timeframes = ['1h']
        self.rest_values = []
        self.cached_values = []
        self.missing_values = []
        self._batch_size = 500
        self._local_connection_timeout = 5
        self._fast_start = fast_start
        self._log_connection = log_connection
        self._force = force
        self._debug = debug

        # rest api configuration
        self._rest_api_host = f'https://{rest_api_prefix}.{NAME}.com'
        self._rest_api_version = 'v1'
        self._rest_api_key = os.environ.get(Envs.REST_API_KEY, rest_api_key)

        # websocket api configuration
        self._host = f'wss://{websocket_api_prefix}.{NAME}.com'
        self._email = os.environ.get(Envs.EMAIL, email)
        self._websocket_api_key = os.environ.get(Envs.WEBSOCKET_API_KEY, websocket_api_key)

        # redis configuration
        self._image_name = f'{NAME}/client:latest'
        self._container_name = NAME
        self._redis_server_address = os.environ.get(Envs.REDIS_SERVER_ADDRESS, '127.0.0.1')
        self._redis_container_port = os.environ.get(Envs.REDIS_SERVER_PORT, 6379)
        self._redis_host_port = os.environ.get(Envs.REDIS_SERVER_PORT, 6379)

        # clients
        self._docker_client = None
        self.redis_client = None

        self._cache_folder = os.environ.get(Envs.CACHE_ROOT, os.path.join(tempfile.gettempdir(), f'{NAME}_cache'))
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(level=logging.DEBUG)

        if not disable_logging:
            self._setup_logger()

        if connect:
            self._start_redis_server()

    @staticmethod
    def _get_cache_keys(route, symbols, timestamps):
        """
        Get all the cache keys.

        :param str route: The data route.
        :param list[str] symbols: A list of symbols.
        :param list[float] timestamps: The timestamps to get.
        :returns: A complete list of all the cache keys.
        :rtype list
        """
        keys = []
        for symbol in symbols:
            keys.extend([f'{NAME}/{route}/{symbol}/{timestamp}' for timestamp in timestamps])
        return keys

    @staticmethod
    def _format_route(route):
        """
        Normalizes the route format.

        :param str route: The data route.
        :returns: A normalizedd the route.
        :rtype str
        """
        return route.strip('/').strip('\\').replace('\\', '/')

    def _get_unused_port_number(self, default_port) -> int:
        """
        Gets an unused port number from the OS.

        :returns: A port number.
        :rtype: int
        """
        if not self._is_port_in_use(default_port):
            return default_port
        else:
            sock = socket.socket()
            sock.bind(('', 0))
            return sock.getsockname()[1]

    def _get_from_cache(self, keys):
        """
        Get the data from the cache that already exists, and which REST requests still need to be made.

        :param list[str] keys: A complete list of all the cache keys.
        :returns: Which REST requests still need to be made.
        :rtype dict
        """
        cache_values = self.redis_client.mget(keys)

        rest_requests = {}
        for key, cache_value in zip(keys, cache_values):
            items = key.split('/')
            symbol = items[-2]
            timestamp = float(items[-1])

            if not cache_value:
                if not rest_requests.get(symbol):
                    rest_requests[symbol] = []
                rest_requests[symbol].append(timestamp)

            # if the data is nothing more than the symbol and id, then this should be
            # marked as missing data.
            elif json.loads(cache_value) == {'id': timestamp, 'symbol': symbol}:
                # if force is true then request the already reported missing values
                if self._force:
                    if not rest_requests.get(symbol):
                        rest_requests[symbol] = []
                    rest_requests[symbol].append(timestamp)
                else:
                    self.missing_values.append(json.loads(cache_value))

            # otherwise, the data is already cached
            else:
                self.cached_values.append(json.loads(cache_value))

        # return the needed rest requests
        return rest_requests

    def _setup_logger(self):
        """
        Sets up the logger.
        """
        formatter = logging.Formatter('%(levelname)s [%(asctime)s] %(name)s: %(message)s', datefmt='%Y-%m-%dT%H:%M:%S')

        # create log handler for console output
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)

        # create log handler for file output
        file_handler = TimedRotatingFileHandler('tickerdax.log', when='midnight')
        file_handler.setFormatter(formatter)

        log_level = logging.INFO
        if self._debug:
            log_level = logging.DEBUG

        logging.basicConfig(
            level=log_level,
            handlers=[console_handler, file_handler]
        )

    def _set_redis_client(self):
        """
        Sets the redis client.
        """
        # verify the connection with the redis server
        try:
            self.redis_client = redis.Redis(
                host=self._redis_server_address,
                port=self._redis_host_port,
                db=0
            )
            self.redis_client.ping()
            if self._log_connection:
                self._logger.info(f'Redis server is connected!')
            return True
        except redis.exceptions.ConnectionError:
            return False

    def _is_port_in_use(self, port: int) -> bool:
        """
        Checks if port number is in use.

        :param int port: A port number.
        :returns: Whether the port is in use.
        :rtype: bool
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as stream:
            return stream.connect_ex((self._redis_server_address, port)) == 0

    def _start_redis_server(self, attempts=0) -> None:
        """
        Starts the redis docker container.
        """
        # quickly try to connect to the running redis server
        if self._fast_start:
            if self._set_redis_client():
                return
            else:
                self._logger.warning(f'Failed to connect with fast start. Starting full reboot...')

        # if in the official docker image, try starting redis manually and
        if os.environ.get(Envs.OFFICIAL_DOCKER_IMAGE) and attempts < 1:
            try:
                subprocess.Popen(
                    f'redis-server /home/redis.conf --port {self._redis_host_port}',
                    shell=True
                )
                self._logger.info('Attempting to start redis server on local system...')
                time.sleep(5)
                if self._set_redis_client():
                    return
            except Exception as error:
                self._logger.debug(error)
                self.report_error(f'Failed to start redis server in {NAME} container')

        # get a unused host port
        self._redis_host_port = self._get_unused_port_number(6379)

        # initialize the docker client
        try:
            self._docker_client = docker.from_env()
        except DockerException:
            self.report_error('Failed to connect to docker. Make sure docker is installed and currently running.')

        # stop any running redis docker containers first
        for container in self._docker_client.containers.list(all=True):
            if self._container_name == container.name:
                if container.status == 'running':
                    self._logger.info(f'Stopping docker container "{self._container_name}"')
                    container.stop()
                    container.wait()
                self._logger.info(f'Removing docker container "{self._container_name}"')
                container.remove()

        # check if the docker image exist locally
        try:
            self._docker_client.images.get(self._image_name)
        except docker.errors.ImageNotFound:
            self._logger.warning(f'Image "{self._image_name}" was not found locally, pulling now...')
            self._docker_client.images.pull(self._image_name)

        # start the tickerdax docker container
        self._logger.info(f'Starting docker container "{self._container_name}"')
        self._docker_client.containers.run(
            name=self._container_name,
            image=self._image_name,
            ports={
                f'{self._redis_container_port}/tcp': (self._redis_server_address, self._redis_host_port)
            },
            volumes=[f'{self._cache_folder}:/data'],
            detach=True
        )

        # try to connect the redis client
        for second in range(self._local_connection_timeout):
            time.sleep(1)
            if self._set_redis_client():
                return
        raise self.report_error('TickerDax failed to connect to the redis server')

    def _batch_request(self, route, symbol, timestamps):
        """
        Batches requests until all timestamps are retrieved.

        :param str route: The data route.
        :param str symbol: The symbol to get.
        :param list[float] timestamps: The timestamps to get.
        :returns A list of all the responses to the request.
        :rtype list[dict]
        """
        result = []
        batch = []
        number_of_timestamp = len(timestamps)
        for index, key in enumerate(timestamps, 1):
            batch.append(key)

            if index % self._batch_size == 0:
                self._logger.info(f'batch requesting {index}/{number_of_timestamp} "{route}/{symbol}" timestamps...')
                result.extend(self._rest_request(route, symbol, batch))

                # clear the batch
                batch.clear()

        # get any remaining items in the last batch
        if batch:
            self._logger.info(
                f'batch requesting {number_of_timestamp}/{number_of_timestamp} "{route}/{symbol}" timestamps...'
            )
            result.extend(self._rest_request(route, symbol, batch))

        self._logger.debug(f'batch "{route}/{symbol}" requests complete!')
        return result

    async def _async_batch_request(self, route, symbol, timestamps):
        """
        Batches requests until all timestamps are retrieved.

        :param str route: The data route.
        :param str symbol: The symbol to get.
        :param list[float] timestamps: The timestamps to get.
        :returns A list of all the responses to the request.
        :rtype list[dict]
        """
        result = []
        batch = []
        number_of_timestamp = len(timestamps)
        for index, key in enumerate(timestamps, 1):
            batch.append(key)

            if index % self._batch_size == 0:
                self._logger.info(f'batch requesting {index}/{number_of_timestamp} "{route}/{symbol}" timestamps...')
                result.extend(self._rest_request(route, symbol, batch))

                # clear the batch
                batch.clear()

        # get any remaining items in the last batch
        if batch:
            self._logger.info(
                f'batch requesting {number_of_timestamp}/{number_of_timestamp} "{route}/{symbol}" timestamps...'
            )
            result.extend(self._rest_request(route, symbol, batch))

        self._logger.debug(f'batch "{route}/{symbol}" requests complete!')
        return result

    async def _stream_to_cache(self, route, symbols):
        """
        Connects to the given route and its symbols and updates the
        cache as it receives new data.

        :param str route: The data route.
        :param list[str] symbols: A list of symbols.
        """
        uri = f'{self._host}?route={route}&symbols={symbols}'
        try:
            async with websockets.connect(
                    uri,
                    extra_headers={'email': self._email, 'token': self._websocket_api_key}
            ) as connected_socket:
                self._logger.info(f'> Connected to {uri}')
                while True:
                    data = json.loads(await connected_socket.recv())
                    symbol = data.get('symbol')
                    timestamp = data.get('id')
                    if symbol and timestamp:
                        for key in self._get_cache_keys(self._format_route(route), [symbol], [timestamp]):
                            self.redis_client.set(key, json.dumps(data))
                            self._logger.info(f'Cached: {pformat(data)}')

        except (ConnectionClosed, ConnectionClosedOK) as error:
            if getattr(error, 'status_code', None) == 1001:
                self._logger.info('refreshing connection...')
                # re-connect if the connection was closed
                await self._stream_to_cache(route, symbols)

        except InvalidStatusCode as error:
            if error.status_code == 401:
                self.report_error(
                    'This email and API key combination are not authorized to connect to '
                    f'the {self._host} websocket API. Please check your credentials.'
                )

    # async def _rest_request(self, route, symbol, timestamps):
    def _rest_request(self, route, symbol, timestamps):
        """
        Preforms a single REST request.

        :param str route: The data route.
        :param str symbol: The symbol to get.
        :param list[float] timestamps: The timestamps to get.
        :return list[dict]: A list of dictionaries as the response to the request.
        """
        try:
            response = requests.get(
                f'{self._rest_api_host}/{self._rest_api_version}/{route}',
                headers={"x-api-key": self._rest_api_key},
                data=json.dumps({
                    'symbol': symbol,
                    'timestamps': timestamps
                })
            )
            if response.ok:
                return response.json()
            else:
                if response.json().get('message') == 'Forbidden':
                    self.report_error(
                        f'This API key is not authorized to connect to the {self._rest_api_host} REST API. '
                        'Please check your credentials.'
                    )
                elif response.json().get('message') == 'Limit Exceeded':
                    self.report_error(
                        f'This API key has exceeded its usage limit. Go to {URL} to upgrade your plan.'
                    )
                else:
                    self.report_error(response.json())
        except Exception as error:
            self.report_error(str(error))

    def _request(self, route, rest_requests):
        """
        A request to first the local cache, then to the REST API if data is missing in the
        cache.

        :param str route: The data route.
        :param dict rest_requests: A dictionary of symbols and timestamps.
        :return list[dict]: A list of dictionaries as the response to the request.
        """
        rest_values = []
        if rest_requests:
            self._logger.info(f'Requesting {route} data from REST API...')
            # gather the symbols concurrently
            for symbol, timestamps in rest_requests.items():
                rest_values.extend(self._batch_request(route, symbol, timestamps))
        return rest_values

    async def _async_request(self, route, rest_requests):
        """
        A request to first the local cache, then to the REST API if data is missing in the
        cache.

        :param str route: The data route.
        :param dict rest_requests: A dictionary of symbols and timestamps.
        :return list[dict]: A list of dictionaries as the response to the request.
        """
        rest_values = []
        if rest_requests:
            self._logger.info(f'Requesting {route} data from REST API...')
            # gather the symbols concurrently
            for result in await asyncio.gather(*[
                self._batch_request(route, symbol, timestamps) for symbol, timestamps in rest_requests.items()
            ]):
                rest_values.extend(result)
        return rest_values

    async def _stream(self, routes):
        """
        Streams all given routes and their symbols concurrently.

        :param dict routes: A dictionary of route names and their symbols.
        """
        await asyncio.gather(*[self._stream_to_cache(
                f'/{self._format_route(route)}', ','.join(symbols)) for route, symbols in routes.items()
        ])

    def _update_cache(self, route, keys, timeframe):
        """
        Saves any new data from the response to the cache.

        :param str route: The data route.
        :param list[str] keys: A complete list of all the cache keys.
        :param str timeframe: The time interval.
        :returns: The combined result of cache values, rest values, and blank values.
        :rtype list
        """
        result = []
        # remove all the cache keys that already had a cached value
        for cached_value in self.cached_values:
            symbol = cached_value.get('symbol')
            timestamp = cached_value.get('id')
            key = f'{NAME}/{route}/{symbol}/{timestamp}'
            if key in keys:
                keys.remove(key)

        result.extend(self.cached_values)

        # cache the rest values
        for rest_value in self.rest_values:
            symbol = rest_value.get('symbol')
            timestamp = rest_value.get('id')
            key = f'{NAME}/{route}/{symbol}/{timestamp}'

            self.redis_client.set(key, json.dumps(rest_value))
            # remove the key now that it is cached
            if key in keys:
                keys.remove(key)
        result.extend(self.rest_values)

        # if there are any remaining keys, then that means they were missing from the rest api
        for key in keys:
            items = key.split('/')
            symbol = items[-2]
            timestamp = float(items[-1])

            missing_value = {'id': timestamp, 'symbol': symbol}
            result.append(missing_value)

            # this will set the missing value in the cache with an expiration time that matches the given timeframe
            self.missing_values.append(missing_value)
            self.redis_client.set(
                key,
                json.dumps(missing_value),
                ex=formatting.convert_timeframe_to_seconds(timeframe)
            )
        return result

    def validate_api_key(self,  key_type):
        """
        Validate whether the key of the given type exists and show and error message.

        :param str key_type: The type of key i.e. REST or WEBSOCKET.
        """
        env_key_name = None
        if key_type == KeyTypes.REST and not self._rest_api_key:
            env_key_name = Envs.REST_API_KEY

        elif key_type == KeyTypes.WEBSOCKET and not self._websocket_api_key:
            env_key_name = Envs.WEBSOCKET_API_KEY

        if env_key_name:
            self.report_error(
                f'The environment variable "{env_key_name}" must be set to your API key from {URL}'
            )

    def get_available_routes(self):
        """
        Gets all available routes from the REST api.

        :returns: A list of all available routes from the REST api.
        :rtype: list
        """
        ignored_routes = ['info/plans']
        routes = []
        for route in requests.get(f'{self._rest_api_host}/openapi.json').json().get('paths', {}).keys():
            route = route.replace(f'/{self._rest_api_version}/', '')
            if route not in ignored_routes:
                routes.append(route)
        return routes

    def get_route(self, route, symbols, start, end, timeframe='1h', asynchronous=False):
        """
        Get data for a route and it's symbols between the start and end times and at the timeframe interval.

        :param str route: The data route.
        :param list[str] symbols: A list of symbols.
        :param datetime start: The UTC start time.
        :param datetime end: The UTC end time.
        :param str timeframe: The time interval.
        :param bool asynchronous: Whether the request is asynchronous.
        :returns: The sorted result.
        :rtype list
        """
        self.cached_values.clear()
        self.rest_values.clear()
        self.missing_values.clear()

        route = self._format_route(route)
        timestamps = formatting.get_timestamp_range(start, end, timeframe)
        keys = self._get_cache_keys(route, symbols, timestamps)

        # get the cached values and determine which rest requests are outstanding
        self._logger.debug(f'Checking "{route}" cache for {symbols}...')
        outstanding_rest_requests = self._get_from_cache(keys)

        # make the request asynchronously or synchronously
        if asynchronous:
            self.rest_values = asyncio.run(self._async_request(route, outstanding_rest_requests))
        else:
            self.rest_values = self._request(route, outstanding_rest_requests)

        result = self._update_cache(route, keys, timeframe)
        return sorted(result, key=lambda i: i['id'])

    def report_error(self, message):
        """
        Reports an error message to the user.

        :param str message: A error message.
        """
        self._logger.error(message)
        sys.exit(1)

    def stream(self, routes):
        """
        Streams all given routes and their symbols to the cache in real-time.

        :param dict routes: A dictionary of route names and their symbols.
        """
        try:
            asyncio.run(self._stream(routes))
        except Exception as error:
            self._logger.error(error)
            self._logger.info('Trying to reconnect...')
            self.stream(routes)


if __name__ == '__main__':
    client = TickerDax()

    pprint(client.get_route(
        route='predictions/ml-order-book-model-1',
        symbols=['BTC'],
        start=datetime.now(tz=timezone.utc) - timedelta(hours=6),
        end=datetime.now(tz=timezone.utc)
    ))

    # client.stream(
    #     routes={
    #         'predictions/ml-order-book-model-1': ['BTC', 'LTC'],
    #     },
    # )



import os
import json
import yaml
import logging
from tickerdax.constants import NAME, Envs
from tickerdax.formatting import truncate_datetime
from datetime import datetime, timezone
from tickerdax.client import TickerDax
from schema import Schema, And, Or, Optional
from pprint import pformat


class ConfigBase:
    """
    The base class that handles the loading and validating of the config.
    """

    def __init__(self, config=None, client_kwargs=None):
        if not client_kwargs:
            client_kwargs = {}

        self.client = TickerDax(**client_kwargs)

        self._logger = logging.getLogger(f'{TickerDax.__name__}.{self.__class__.__name__}')
        self._logger.setLevel(logging.INFO)

        self._config_file_path = os.environ.get(Envs.CONFIG, config)
        self._config = self._get_config_data()
        self._validate()

        self._timeframe = self._config.get('timeframe')
        self._now = datetime.now(tz=timezone.utc)
        self._start, self._end = self._get_time_range()

    def _validate(self):
        """
        Validates the class attributes.
        """
        timeframes = self.client.supported_timeframes
        schema = Schema({
            'routes': And(dict),
            'timeframe': And(str, lambda s: s in timeframes,
                             error=f'Your configs timeframe is not one of the supported timeframes {timeframes}'),
            'start': Or(str, datetime),
            Optional('end'): Or(str, datetime),
            Optional('fill_to_now'): And(bool),
            Optional('force'): And(bool)
        })
        schema.validate(self._config)

        available_routes = self.client.get_available_routes()
        for route in self._config.get('routes', {}).keys():
            if route not in available_routes:
                self.client.report_error(
                    f'The route "{route}" is not valid. Possible routes are: {pformat(available_routes)}'
                )

    def _get_config_data(self):
        """
        Gets the data from the config file.

        :return dict: A dictionary of config data.
        """
        if os.path.exists(self._config_file_path):
            _, extension = os.path.splitext(self._config_file_path)
            with open(self._config_file_path, 'r') as config_file:
                if extension == '.json':
                    return json.load(config_file).get(NAME, {})
                elif extension in ['.yaml', '.yml']:
                    return yaml.safe_load(config_file).get(NAME, {})
        else:
            raise self.client.report_error(f'The config file "{self._config_file_path}" does not exist on disk')

    def _get_time_range(self):
        """
        Gets the UTC start and end times for the data download.

        :return tuple(datetime, datetime):
        """
        start_string = self._config.get('start')
        end_string = self._config.get('end')

        if isinstance(start_string, str):
            start = datetime.strptime(start_string, '%Y-%m-%dT%H:%M:%S')
        else:
            start = start_string

        # if there is no ending time the current time is used
        if end_string:
            if isinstance(start_string, str):
                end = datetime.strptime(end_string, '%Y-%m-%dT%H:%M:%S')
            else:
                end = end_string
        else:
            end = self._now

        return truncate_datetime(start, self._timeframe), truncate_datetime(end, self._timeframe)



from tickerdax.constants import KeyTypes
from tickerdax.config_base import ConfigBase
from tickerdax.formatting import truncate_datetime, convert_timeframe_to_seconds, show_download_summary


class Downloader(ConfigBase):
    """
    Downloads data from the tickerdax.com data api using a provided config.
    """
    def __init__(self, config, client_kwargs=None, till_now=False):
        super(Downloader, self).__init__(config, client_kwargs)
        self._missing_data = {}

        # forces the end to be now
        if till_now:
            self._end = truncate_datetime(self._now, self._timeframe)

        self.download()

    def _validate(self):
        super(Downloader, self)._validate()
        self.client.validate_api_key(KeyTypes.REST)

    def track_missing_ranges(self, route):
        self.client.missing_values.sort(key=lambda i: i['id'])
        for missing_value in self.client.missing_values:
            path = f"{route}/{missing_value['symbol']}"
            if not self._missing_data.get(path):
                self._missing_data[path] = []

            if missing_value not in self._missing_data[path]:
                self._missing_data[path].append(missing_value)

    def get_missing_ranges(self):
        ranges = {}
        for route, missing_values in self._missing_data.items():
            ranges[route] = []

            last_missing_timestamp = None
            gap_start_timestamp = None

            for missing_value in missing_values:
                if not gap_start_timestamp:
                    gap_start_timestamp = missing_value['id']

                if last_missing_timestamp:
                    incremented_timestamp = last_missing_timestamp + convert_timeframe_to_seconds(self._timeframe)
                    if missing_value['id'] != incremented_timestamp:
                        ranges[route].append([gap_start_timestamp, last_missing_timestamp])
                        gap_start_timestamp = None

                last_missing_timestamp = missing_value['id']
            ranges[route].append([gap_start_timestamp, last_missing_timestamp])

        return ranges

    def download(self):
        """
        Downloads data from the algo trading REST API base on the bot config files.
        """
        downloaded_items = 0
        cached_items = 0
        missing_items = 0

        routes = self._config.get('routes', {})
        if routes:
            for route, symbols in routes.items():
                self._logger.info(f'Downloading {route} history from "{self._start}" to "{self._end}"...')
                self.client.get_route(
                    route=route,
                    symbols=symbols,
                    start=self._start,
                    end=self._end,
                    timeframe=self._timeframe
                )
                self._logger.info(f'Completed {len(self.client.rest_values)} "{route}" downloads')
                cached_items += len(self.client.cached_values)
                downloaded_items += len(self.client.rest_values)
                missing_items += len(self.client.missing_values)
                self.track_missing_ranges(route)

            if downloaded_items:
                self._logger.info('Saving download to disk..')
                self.client.redis_client.bgsave()

            self._logger.info('Preparing summary...')
            show_download_summary(
                cached_items=cached_items,
                downloaded_items=downloaded_items,
                missing_items=missing_items,
                missing_ranges=self.get_missing_ranges()
            )


import calendar
from rich.table import Table
from rich.tree import Tree
from rich.panel import Panel
from rich import print
from datetime import datetime, timezone
from tickerdax.constants import URL


def truncate_datetime(date, timeframe):
    kwargs = {
        'year': date.year
    }
    if timeframe.endswith(('M', 'd', 'h', 'm')):
        kwargs['month'] = date.month

    if timeframe.endswith(('d', 'h', 'm')):
        kwargs['day'] = date.day

    if timeframe.endswith(('h', 'm')):
        kwargs['hour'] = date.hour

    if timeframe.endswith('m'):
        kwargs['minute'] = date.minute

    return datetime(
        tzinfo=timezone.utc,
        **kwargs
    )


def get_unix_time(date, timeframe):
    return float(calendar.timegm(truncate_datetime(date, timeframe).timetuple()))


def get_timestamp_range(start, end, timeframe):
    """
    Gets a range of timestamps between the start and end dates.

    :param datetime start: A start time.
    :param datetime end: A end time.
    :param str timeframe: The string value of the timeframe, i.e. 1m, 1h, 1d, etc.
    :rtype: list[int]
    """
    start = get_unix_time(start, timeframe)
    end = get_unix_time(end, timeframe)
    range_seconds = end - start
    timeframe_in_seconds = convert_timeframe_to_seconds(timeframe)
    time_intervals = range_seconds / timeframe_in_seconds

    timestamps = []
    for time_interval in range(int(time_intervals) + 1):
        timestamps.append(start + (time_interval * timeframe_in_seconds))

    return timestamps


def convert_timeframe_to_seconds(timeframe):
    seconds_per_unit = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}
    return int(timeframe[:-1]) * seconds_per_unit[timeframe[-1]]


def show_download_summary(cached_items, downloaded_items, missing_items, missing_ranges):
    tree = Tree(f'- {cached_items} of the requested items were already cached.\n'
                f'- {downloaded_items} items were downloaded.\n'
                f'- {missing_items} items are missing.',)
    for key, values in missing_ranges.items():
        table = Table(title=key)
        table.add_column("Begins", style="cyan")
        table.add_column("Ends", style="green")

        for start, end in values:
            if start and end:
                table.add_row(
                    datetime.fromtimestamp(start).strftime('%Y-%m-%dT%H:%M:%S'),
                    datetime.fromtimestamp(end).strftime('%Y-%m-%dT%H:%M:%S')
                )
        tree.add(table)
    print(Panel(tree, title="Download Summary"))


def show_routes(routes):
    table = Table(title=f'All Available Routes from {URL}')
    table.add_column("Routes", style="cyan")
    table.add_column("Timeframes", style="green")
    for route in routes:
        table.add_row(
            route,
            ','.join(['1h'])
        )
    print()
    print(table)



import os
import typer
import shutil
import pkg_resources
import tickerdax
import tomlkit
import art
from pathlib import Path
from rich import print
from rich.text import Text
from typing import Optional
from tickerdax import formatting
from tickerdax.constants import Envs
from tickerdax.client import TickerDax
from tickerdax.streamer import Streamer
from tickerdax.downloader import Downloader
from typer import rich_utils
from dotenv import load_dotenv

# load in the env file if there is one
load_dotenv()

# monkey patch this function, so we can add in header ascii art
rich_format_help = rich_utils.rich_format_help


def override_help(*args, **kwargs):
    text = Text(art.text2art('TICKERDAX'))
    text.stylize("bold blue")
    print(text)
    rich_format_help(*args, **kwargs)


rich_utils.rich_format_help = override_help

# create cli app
app = typer.Typer(
    pretty_exceptions_show_locals=False
)


def validate_callback(value):
    if not value:
        raise typer.BadParameter('This must be set')


config_argument = typer.Argument(
    'config.yaml',
    exists=True,
    file_okay=True,
    dir_okay=False,
    writable=False,
    readable=True,
    resolve_path=True,
    show_default=False,
    help="Path to config file.",
    envvar=Envs.CONFIG
)
rest_api_key_argument = typer.Argument(
    None,
    callback=validate_callback,
    show_default=False,
    help="Your REST api key.",
    envvar=Envs.REST_API_KEY
)

websocket_api_key_argument = typer.Argument(
    None,
    callback=validate_callback,
    show_default=False,
    help="Your websocket api key.",
    envvar=Envs.WEBSOCKET_API_KEY
)

force_argument = typer.Option(
    False,
    help="Forces new REST requests for all missing data, even if that data has already been marked as missing"
)


def version_callback(value: bool):
    """
    Shows the current cli version

    :param bool value: Whether the version flag was passed.
    """
    if value:
        project_file_path = os.path.join(os.path.dirname(__file__), os.path.pardir, 'pyproject.toml')
        if os.path.exists(project_file_path):
            with open(project_file_path, "rb") as project_file:
                data = tomlkit.load(project_file)
                version = data.get('tool', {}).get('poetry', {}).get('version', '0.0.1')
        else:
            version = pkg_resources.get_distribution(tickerdax.__name__).version
        print(f"TickerDax CLI Version: {version}")
        raise typer.Exit()


@app.callback(no_args_is_help=True)
def callback(version: Optional[bool] = typer.Option(None, "--version", callback=version_callback)):
    """
    The TickerDax CLI tool interfaces with the tickerdax.com REST and websockets APIs. It
    handles common data operations like batch downloading, streaming, and caching data
    locally to minimize network requests.
    """


@app.command()
def create_config():
    """
    Creates a new ticker dax config.
    """
    file_formats = ['json', 'yaml']
    extension = typer.prompt(
        'What file format do you want to use? (json or yaml)',
        default='yaml'
    )
    if extension not in file_formats:
        raise typer.BadParameter(f'Your choice {extension} is not one of the valid formats {file_formats}')

    # copy the example config to the current working directory
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), 'example_configs', f'config.{extension}'),
        os.path.join(os.getcwd(), f'config.{extension}')
    )


@app.command()
def list_routes():
    """
    Lists all routes available to download or stream.
    """
    tickerdax_client = TickerDax(connect=False)
    routes = tickerdax_client.get_available_routes()
    formatting.show_routes(routes)


@app.command()
def download(
        config: Optional[Path] = config_argument,
        rest_api_key: str = rest_api_key_argument,
        force: bool = force_argument,
):
    """
    Downloads data from the routes with the time interval specified in your config.
    """
    Downloader(
        config=config,
        client_kwargs={
            'rest_api_key': rest_api_key,
            'force': force,
        }
    )


@app.command()
def stream(
        config: Optional[Path] = config_argument,
        rest_api_key: str = rest_api_key_argument,
        websocket_api_key: str = websocket_api_key_argument,
        force: bool = force_argument
):
    """
    Streams data from the routes specified in your config.
    """
    Streamer(
        config=config,
        client_kwargs={
            'rest_api_key': rest_api_key,
            'websocket_api_key': websocket_api_key,
            'force': force,
        }
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tickerdax']

package_data = \
{'': ['*'], 'tickerdax': ['example_configs/*']}

install_requires = \
['art>=5.8,<6.0',
 'docker>=6.0.1,<7.0.0',
 'fastapi>=0.88.0,<0.89.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'pyyaml>=6.0,<7.0',
 'redis>=4.4.0,<5.0.0',
 'schema>=0.7.5,<0.8.0',
 'tomlkit>=0.11.6,<0.12.0',
 'typer[all]>=0.7.0,<0.8.0',
 'uvicorn>=0.20.0,<0.21.0',
 'websockets>=10.4,<11.0']

entry_points = \
{'console_scripts': ['tickerdax = tickerdax.cli:app']}

setup_kwargs = {
    'name': 'tickerdax',
    'version': '0.0.26',
    'description': 'A python client for tickerdax.com with a built-in caching system.',
    'long_description': '<p align="center">\n  <img width="200" src="https://tickerdax.com/assets/images/logo/logo.svg" alt="icon"/>\n</p>\n<h1 align="center">TickerDax Client</h1>\n<br></br>\n\n[![CI](https://github.com/TickerDax/tickerdax-client/actions/workflows/ci.yaml/badge.svg)](https://github.com/TickerDax/tickerdax-client/actions/workflows/ci.yaml)\n\nA python package that interfaces with the tickerdax.com REST and websockets API. It handles common data operations\nlike batch downloading data, streaming real-time data, and caching data locally to minimize network requests.\n\n## Installation\nYou can install this package with pip by running the command below.\n```shell\npip install tickerdax\n```\n\n## Docker Dependency\nThis client interfaces with a redis docker container. In order for the package to work, you must first install\ndocker. Here are instructions per platform.\n### Mac\n[Instructions](https://docs.docker.com/desktop/install/mac-install/)\n### Linux\n[Instructions](https://docs.docker.com/desktop/install/linux-install/)\n### Windows\nNote on windows you must first install [WSL](https://docs.microsoft.com/en-us/windows/wsl/install) then you can install docker.\n[Instructions](https://docs.docker.com/desktop/install/windows-install/)\n\n## Python Examples\nHere is a basic example of getting historical data using the python SDK.\n### Get historical data\n```python\nfrom pprint import pprint\nfrom datetime import datetime, timezone\nfrom tickerdax.client import TickerDax\n\nclient = TickerDax()\npprint(client.get_route(\n    route=\'predictions/ml-order-book-model-1\',\n    symbols=["BTC"],\n    start=datetime.now(tz=timezone.utc),\n    end=datetime.now(tz=timezone.utc)\n))\n```\nNote that if this data doesn\'t exist in your cache, the data will be fetched from the REST API. All \nsubsequent calls to the same data will only be from the cache and not the REST API.\nThis is designed give you lighting fast responses and ultimately deliver data to you a cheaper cost.\n\n### Stream realtime data\nThis is how you can stream data to your cache. This will run indefinitely and fill\nyour local cache as new data is available.\n```python\nclient.stream(\n    routes={\n        \'predictions/ml-order-book-model-1\': [\'BTC\', \'LTC\'],\n    },\n)\n```\nIn another process you can call `client.get_route()` as many times you like or whenever your\napp re-evaluates. The data will be available once it is updated by the stream.\n\n\n## CLI\nThe package also has a command line interface.\n```text\nUsage: tickerdax [OPTIONS] COMMAND [ARGS]...                                \n                                                                            \n  TickerDax version 0.0.21. A CLI tool that interfaces with the tickerdax.com\n  REST and websockets API. It handles common data operations like batch     \n  downloading, streaming, and caching data locally to minimize network      \n  requests.                                                                 \n                                                                            \nOptions:                                                                    \n  --help  Show this message and exit.\n\nCommands:\n  create-config\n  download\n  list-routes\n  stream\n```\n\n## Environment Variables\n| Name                             | Description                                                                                                                                |\n|----------------------------------|--------------------------------------------------------------------------------------------------------------------------------------------|\n| `TICKERDAX_CONFIG`               | A file path to the config file for the CLI.                                                                                                |\n| `TICKERDAX_EMAIL`                | Your email linked to your tickerdax.com account.                                                                                           |\n| `TICKERDAX_REST_API_KEY`         | Your REST API created with your tickerdax.com account.                                                                                     |\n| `TICKERDAX_WEBSOCKET_API_KEY`    | Your websocket API created with your tickerdax.com account.                                                                                |\n| `TICKERDAX_CACHE_ROOT`           | An alternative persistent cache location on disk. By default this is written into a `tickerdax_cache` folder in your system\'s temp folder. |\n| `TICKERDAX_REDIS_SERVER_ADDRESS` | An alternative redis server address. Can be useful if redis is on another address besides localhost.                                       |\n| `TICKERDAX_REDIS_SERVER_PORT`    | An alternative redis server port. Can be useful if redis needs to user another port besides 6379.                                          |',
    'author': 'tickerdax.com',
    'author_email': 'info@tickerdax.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tickerdax/tickerdax-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0.0',
}


setup(**setup_kwargs)
